@@ -69,7 +69,7 @@ static int hnbgw_tx_hnb_register_rej(struct hnb_context *ctx)
 		return rc;
 	}
 
-	/* generate a successfull outcome PDU */
+	/* generate a unsuccessful outcome PDU */
 	msg = hnbap_generate_unsuccessful_outcome(HNBAP_ProcedureCode_id_HNBRegister,
 						  HNBAP_Criticality_reject,
 						  &asn_DEF_HNBAP_HNBRegisterReject,
@@ -111,7 +111,7 @@ static int hnbgw_tx_hnb_register_acc(struct hnb_context *ctx)
 		return rc;
 	}
 
-	/* generate a successfull outcome PDU */
+	/* generate a successful outcome PDU */
 	msg = hnbap_generate_successful_outcome(HNBAP_ProcedureCode_id_HNBRegister,
 					       HNBAP_Criticality_reject,
 					       &asn_DEF_HNBAP_HNBRegisterAccept,
@@ -284,6 +284,7 @@ static int hnbgw_tx_ue_register_acc_tmsi(struct hnb_context *hnb, HNBAP_UE_Ident
 	uint32_t ctx_id;
 	uint32_t tmsi = 0;
 	struct ue_context *ue;
+	struct ue_context *ue_allocated = NULL;
 	int rc;
 
 	memset(&accept, 0, sizeof(accept));
@@ -331,14 +332,19 @@ static int hnbgw_tx_ue_register_acc_tmsi(struct hnb_context *hnb, HNBAP_UE_Ident
 
 	ue = ue_context_by_tmsi(hnb->gw, tmsi);
 	if (!ue)
-		ue = ue_context_alloc(hnb, NULL, tmsi);
+		ue = ue_allocated = ue_context_alloc(hnb, NULL, tmsi);
 
 	asn1_u24_to_bitstring(&accept.context_ID, &ctx_id, ue->context_id);
 
 	memset(&accept_out, 0, sizeof(accept_out));
 	rc = hnbap_encode_ueregisteraccepties(&accept_out, &accept);
-	if (rc < 0)
+	if (rc < 0) {
+		/* If we allocated the UE context but the UE REGISTER fails, get rid of it again: there will likely
+		 * never be a UE DE-REGISTER for this UE from the HNB, and the ue_context would linger forever. */
+		if (ue_allocated)
+			ue_context_free(ue_allocated);
 		return rc;
+	}
 
 	msg = hnbap_generate_successful_outcome(HNBAP_ProcedureCode_id_UERegister,
 						HNBAP_Criticality_reject,
@@ -388,22 +394,26 @@ static int hnbgw_rx_hnb_deregister(struct hnb_context *ctx, ANY_t *in)
 	LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "HNB-DE-REGISTER cause=%s\n", hnbap_cause_str(&ies.cause));
 
 	hnbap_free_hnbde_registeries(&ies);
-	hnb_context_release(ctx);
+	ctx->hnb_registered = false;
 
 	return 0;
 }
 
 static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 {
-	struct hnb_context *hnb;
+	struct hnb_context *hnb, *tmp;
 	HNBAP_HNBRegisterRequestIEs_t ies;
 	int rc;
 	struct osmo_plmn_id plmn;
+	struct osmo_fd *ofd = osmo_stream_srv_get_ofd(ctx->conn);
+	char name[OSMO_SOCK_NAME_MAXLEN];
+
+	osmo_sock_get_name_buf(name, sizeof(name), ofd->fd);
 
 	rc = hnbap_decode_hnbregisterrequesties(&ies, in);
 	if (rc < 0) {
-		LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Failure to decode HNB-REGISTER-REQ from %s: rc=%d\n",
-			ctx->identity_info, rc);
+		LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Failure to decode HNB-REGISTER-REQ %s from %s: rc=%d\n",
+		       ctx->identity_info, name, rc);
 		return rc;
 	}
 
@@ -418,21 +428,48 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 	ctx->id.mcc = plmn.mcc;
 	ctx->id.mnc = plmn.mnc;
 
-	llist_for_each_entry(hnb, &ctx->gw->hnb_list, list) {
+	llist_for_each_entry_safe(hnb, tmp, &ctx->gw->hnb_list, list) {
 		if (hnb->hnb_registered && ctx != hnb && memcmp(&ctx->id, &hnb->id, sizeof(ctx->id)) == 0) {
-			struct osmo_fd *ofd = osmo_stream_srv_get_ofd(ctx->conn);
-			char *name = osmo_sock_get_name(ctx, ofd->fd);
+			/* If it's coming from the same remote IP addr+port, then it must be our internal
+			 * fault (bug), and we release the old context to keep going... */
+			struct osmo_fd *other_fd = osmo_stream_srv_get_ofd(hnb->conn);
+			struct osmo_sockaddr other_osa = {};
+			struct osmo_sockaddr cur_osa = {};
+			socklen_t len = sizeof(other_osa);
+			if (getpeername(other_fd->fd, &other_osa.u.sa, &len) < 0) {
+				LOGHNB(ctx, DHNBAP, LOGL_ERROR, "BUG! Found old registered HNB with invalid socket, releasing it\n");
+				hnb_context_release(hnb);
+				continue;
+			}
+			len = sizeof(cur_osa);
+			if (getpeername(ofd->fd, &cur_osa.u.sa, &len) < 0) {
+				LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Error getpeername(): %s\n", strerror(errno));
+				if (osmo_sockaddr_cmp(&cur_osa, &other_osa) == 0) {
+					LOGHNB(ctx, DHNBAP, LOGL_ERROR, "BUG! Found old registered HNB with same remote address, releasing it\n");
+					hnb_context_release(hnb);
+					continue;
+				}
+			} else if (osmo_sockaddr_cmp(&cur_osa, &other_osa) == 0) {
+				LOGHNB(ctx, DHNBAP, LOGL_ERROR, "BUG! Found old registered HNB with same remote address, releasing it\n");
+				hnb_context_release(hnb);
+				continue;
+			} /* else: addresses are different, we continue below */
+
+			/* If new conn registering same HNB is from anoter remote addr+port, let's reject it to avoid
+			 * misconfigurations or someone trying to impersonate an already working HNB: */
 			LOGHNB(ctx, DHNBAP, LOGL_ERROR, "rejecting HNB-REGISTER-REQ with duplicate cell identity "
 				"MCC=%u,MNC=%u,LAC=%u,RAC=%u,SAC=%u,CID=%u from %s\n",
 				ctx->id.mcc, ctx->id.mnc, ctx->id.lac, ctx->id.rac, ctx->id.sac, ctx->id.cid, name);
-			talloc_free(name);
+			hnbap_free_hnbregisterrequesties(&ies);
 			return hnbgw_tx_hnb_register_rej(ctx);
 		}
 	}
 
-	ctx->hnb_registered = true;
+	LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "HNB-REGISTER-REQ %s MCC=%u,MNC=%u,LAC=%u,RAC=%u,SAC=%u,CID=%u from %s%s\n",
+	       ctx->identity_info, ctx->id.mcc, ctx->id.mnc, ctx->id.lac, ctx->id.rac, ctx->id.sac, ctx->id.cid,
+	       name, ctx->hnb_registered ? " (duplicated)" : "");
 
-	LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "HNB-REGISTER-REQ from %s\n", ctx->identity_info);
+	ctx->hnb_registered = true;
 
 	/* Send HNBRegisterAccept */
 	rc = hnbgw_tx_hnb_register_acc(ctx);
@@ -444,6 +481,7 @@ static int hnbgw_rx_ue_register_req(struct hnb_context *ctx, ANY_t *in)
 {
 	HNBAP_UERegisterRequestIEs_t ies;
 	struct ue_context *ue;
+	struct ue_context *ue_allocated = NULL;
 	char imsi[16];
 	int rc;
 
@@ -485,11 +523,18 @@ static int hnbgw_rx_ue_register_req(struct hnb_context *ctx, ANY_t *in)
 
 	ue = ue_context_by_imsi(ctx->gw, imsi);
 	if (!ue)
-		ue = ue_context_alloc(ctx, imsi, 0);
+		ue = ue_allocated = ue_context_alloc(ctx, imsi, 0);
 
 	hnbap_free_ueregisterrequesties(&ies);
 	/* Send UERegisterAccept */
-	return hnbgw_tx_ue_register_acc(ue);
+	rc = hnbgw_tx_ue_register_acc(ue);
+	if (rc < 0) {
+		/* If we allocated the UE context but the UE REGISTER fails, get rid of it again: there will likely
+		 * never be a UE DE-REGISTER for this UE from the HNB, and the ue_context would linger forever. */
+		if (ue_allocated)
+			ue_context_free(ue_allocated);
+	}
+	return rc;
 }
 
 static int hnbgw_rx_ue_deregister(struct hnb_context *ctx, ANY_t *in)
@@ -534,32 +579,47 @@ static int hnbgw_rx_initiating_msg(struct hnb_context *hnb, HNBAP_InitiatingMess
 {
 	int rc = 0;
 
-	switch (imsg->procedureCode) {
-	case HNBAP_ProcedureCode_id_HNBRegister:	/* 8.2 */
-		rc = hnbgw_rx_hnb_register_req(hnb, &imsg->value);
-		break;
-	case HNBAP_ProcedureCode_id_HNBDe_Register:	/* 8.3 */
-		rc = hnbgw_rx_hnb_deregister(hnb, &imsg->value);
-		break;
-	case HNBAP_ProcedureCode_id_UERegister: 	/* 8.4 */
-		rc = hnbgw_rx_ue_register_req(hnb, &imsg->value);
-		break;
-	case HNBAP_ProcedureCode_id_UEDe_Register:	/* 8.5 */
-		rc = hnbgw_rx_ue_deregister(hnb, &imsg->value);
-		break;
-	case HNBAP_ProcedureCode_id_ErrorIndication:	/* 8.6 */
-		rc = hnbgw_rx_err_ind(hnb, &imsg->value);
-		break;
-	case HNBAP_ProcedureCode_id_TNLUpdate:	/* 8.9 */
-	case HNBAP_ProcedureCode_id_HNBConfigTransfer:	/* 8.10 */
-	case HNBAP_ProcedureCode_id_RelocationComplete:	/* 8.11 */
-	case HNBAP_ProcedureCode_id_U_RNTIQuery:	/* 8.12 */
-	case HNBAP_ProcedureCode_id_privateMessage:
-		LOGHNB(hnb, DHNBAP, LOGL_NOTICE, "Unimplemented HNBAP Procedure %ld\n", imsg->procedureCode);
-		break;
-	default:
-		LOGHNB(hnb, DHNBAP, LOGL_NOTICE, "Unknown HNBAP Procedure %ld\n", imsg->procedureCode);
-		break;
+	if (!hnb->hnb_registered) {
+		switch (imsg->procedureCode) {
+		case HNBAP_ProcedureCode_id_HNBRegister:	/* 8.2 */
+			rc = hnbgw_rx_hnb_register_req(hnb, &imsg->value);
+			break;
+		case HNBAP_ProcedureCode_id_HNBDe_Register:	/* 8.3 */
+			rc = hnbgw_rx_hnb_deregister(hnb, &imsg->value);
+			break;
+		default:
+			LOGHNB(hnb, DHNBAP, LOGL_NOTICE, "HNBAP Procedure %ld not permitted for de-registered HNB\n",
+				imsg->procedureCode);
+			break;
+		}
+	} else {
+		switch (imsg->procedureCode) {
+		case HNBAP_ProcedureCode_id_HNBRegister:	/*  8.2.4: Abnormal Condition, Accept. */
+			rc = hnbgw_rx_hnb_register_req(hnb, &imsg->value);
+			break;
+		case HNBAP_ProcedureCode_id_HNBDe_Register:	/* 8.3 */
+			rc = hnbgw_rx_hnb_deregister(hnb, &imsg->value);
+			break;
+		case HNBAP_ProcedureCode_id_UERegister:		/* 8.4 */
+			rc = hnbgw_rx_ue_register_req(hnb, &imsg->value);
+			break;
+		case HNBAP_ProcedureCode_id_UEDe_Register:	/* 8.5 */
+			rc = hnbgw_rx_ue_deregister(hnb, &imsg->value);
+			break;
+		case HNBAP_ProcedureCode_id_ErrorIndication:	/* 8.6 */
+			rc = hnbgw_rx_err_ind(hnb, &imsg->value);
+			break;
+		case HNBAP_ProcedureCode_id_TNLUpdate:		/* 8.9 */
+		case HNBAP_ProcedureCode_id_HNBConfigTransfer:	/* 8.10 */
+		case HNBAP_ProcedureCode_id_RelocationComplete:	/* 8.11 */
+		case HNBAP_ProcedureCode_id_U_RNTIQuery:	/* 8.12 */
+		case HNBAP_ProcedureCode_id_privateMessage:
+			LOGHNB(hnb, DHNBAP, LOGL_NOTICE, "Unimplemented HNBAP Procedure %ld\n", imsg->procedureCode);
+			break;
+		default:
+			LOGHNB(hnb, DHNBAP, LOGL_NOTICE, "Unknown HNBAP Procedure %ld\n", imsg->procedureCode);
+			break;
+		}
 	}
 
 	return rc;
